Index: ext/standard/array.c
===================================================================
--- ext/standard/array.c	(revision 35093)
+++ ext/standard/array.c	(revision 35401)
@@ -1653,24 +1653,28 @@
 		high = (unsigned char *)Z_STRVAL_P(zhigh);
 
 		if (*low > *high) {		/* Negative Steps */
+			unsigned char ch = *low;
+
 			if (lstep <= 0) {
 				err = 1;
 				goto err;
 			}
-			for (; *low >= *high; (*low) -= (unsigned int)lstep) {
-				add_next_index_stringl(return_value, (const char *)low, 1, 1);
-				if (((signed int)*low - lstep) < 0) {
+			for (; ch >= *high; ch -= (unsigned int)lstep) {
+				add_next_index_stringl(return_value, (const char *)&ch, 1, 1);
+				if (((signed int)ch - lstep) < 0) {
 					break;
 				}
 			}
 		} else if (*high > *low) {	/* Positive Steps */
+			unsigned char ch = *low;
+
 			if (lstep <= 0) {
 				err = 1;
 				goto err;
 			}
-			for (; *low <= *high; (*low) += (unsigned int)lstep) {
-				add_next_index_stringl(return_value, (const char *)low, 1, 1);
-				if (((signed int)*low + lstep) > 255) {
+			for (; ch <= *high; ch += (unsigned int)lstep) {
+				add_next_index_stringl(return_value, (const char *)&ch, 1, 1);
+				if (((signed int)ch + lstep) > 255) {
 					break;
 				}
 			}
Index: ext/standard/string.c
===================================================================
--- ext/standard/string.c	(revision 35093)
+++ ext/standard/string.c	(revision 35401)
@@ -3589,7 +3589,9 @@
 														   replace_value, replace_len, &Z_STRLEN(temp_result), case_sensitivity, replace_count);
 			}
 
-			efree(Z_STRVAL_P(result));
+			if (!IS_INTERNED(Z_STRVAL_P(result))) {
+				efree(Z_STRVAL_P(result));
+			}
 			Z_STRVAL_P(result) = Z_STRVAL(temp_result);
 			Z_STRLEN_P(result) = Z_STRLEN(temp_result);
 
@@ -4244,6 +4246,7 @@
 	char *tbuf, *buf, *p, *tp, *rp, c, lc;
 	int br, i=0, depth=0, in_q = 0;
 	int state = 0, pos;
+	char *allow_free;
 
 	if (stateptr)
 		state = *stateptr;
@@ -4255,7 +4258,12 @@
 	rp = rbuf;
 	br = 0;
 	if (allow) {
-		php_strtolower(allow, allow_len);
+		if (IS_INTERNED(allow)) {
+			allow_free = allow = zend_str_tolower_dup(allow, allow_len);
+		} else {
+			allow_free = NULL;
+			php_strtolower(allow, allow_len);
+		}
 		tbuf = emalloc(PHP_TAG_BUF_SIZE + 1);
 		tp = tbuf;
 	} else {
@@ -4494,8 +4502,12 @@
 		*rp = '\0';
 	}
 	efree(buf);
-	if (allow)
+	if (allow) {
 		efree(tbuf);
+		if (allow_free) {
+			efree(allow_free);
+		}
+	}
 	if (stateptr)
 		*stateptr = state;
 
Index: ext/phar/phar.c
===================================================================
--- ext/phar/phar.c	(revision 297868)
+++ ext/phar/phar.c	(working copy)
@@ -1736,32 +1736,31 @@
 static int phar_analyze_path(const char *fname, const char *ext, int ext_len, int for_create TSRMLS_DC) /* {{{ */
 {
 	php_stream_statbuf ssb;
-	char *realpath, old, *a = (char *)(ext + ext_len);
+	char *realpath;
+	char *filename = estrndup(fname, (ext - fname) + ext_len);
 
-	old = *a;
-	*a = '\0';
-
-	if ((realpath = expand_filepath(fname, NULL TSRMLS_CC))) {
+	if ((realpath = expand_filepath(filename, NULL TSRMLS_CC))) {
 #ifdef PHP_WIN32
 		phar_unixify_path_separators(realpath, strlen(realpath));
 #endif
 		if (zend_hash_exists(&(PHAR_GLOBALS->phar_fname_map), realpath, strlen(realpath))) {
-			*a = old;
 			efree(realpath);
+			efree(filename);
 			return SUCCESS;
 		}
 
 		if (PHAR_G(manifest_cached) && zend_hash_exists(&cached_phars, realpath, strlen(realpath))) {
-			*a = old;
 			efree(realpath);
+			efree(filename);
 			return SUCCESS;
 		}
 		efree(realpath);
 	}
 
-	if (SUCCESS == php_stream_stat_path((char *) fname, &ssb)) {
-		*a = old;
+	if (SUCCESS == php_stream_stat_path((char *) filename, &ssb)) {
 
+		efree(filename);
+
 		if (ssb.sb.st_mode & S_IFDIR) {
 			return FAILURE;
 		}
@@ -1775,57 +1774,56 @@
 		char *slash;
 
 		if (!for_create) {
-			*a = old;
+			efree(filename);
 			return FAILURE;
 		}
 
-		slash = (char *) strrchr(fname, '/');
-		*a = old;
+		slash = (char *) strrchr(filename, '/');
 
 		if (slash) {
-			old = *slash;
 			*slash = '\0';
 		}
 
-		if (SUCCESS != php_stream_stat_path((char *) fname, &ssb)) {
-			if (slash) {
-				*slash = old;
-			} else {
-				if (!(realpath = expand_filepath(fname, NULL TSRMLS_CC))) {
+		if (SUCCESS != php_stream_stat_path((char *) filename, &ssb)) {
+			if (!slash) {
+				if (!(realpath = expand_filepath(filename, NULL TSRMLS_CC))) {
+					efree(filename);
 					return FAILURE;
 				}
 #ifdef PHP_WIN32
 				phar_unixify_path_separators(realpath, strlen(realpath));
 #endif
-				a = strstr(realpath, fname) + ((ext - fname) + ext_len);
-				*a = '\0';
+				slash = strstr(realpath, filename) + ((ext - fname) + ext_len);
+				*slash = '\0';
 				slash = strrchr(realpath, '/');
 
 				if (slash) {
 					*slash = '\0';
 				} else {
 					efree(realpath);
+					efree(filename);
 					return FAILURE;
 				}
 
 				if (SUCCESS != php_stream_stat_path(realpath, &ssb)) {
 					efree(realpath);
+					efree(filename);
 					return FAILURE;
 				}
 
 				efree(realpath);
 
 				if (ssb.sb.st_mode & S_IFDIR) {
+					efree(filename);
 					return SUCCESS;
 				}
 			}
 
+			efree(filename);
 			return FAILURE;
 		}
 
-		if (slash) {
-			*slash = old;
-		}
+		efree(filename);
 
 		if (ssb.sb.st_mode & S_IFDIR) {
 			return SUCCESS;
Index: ext/filter/sanitizing_filters.c
===================================================================
--- ext/filter/sanitizing_filters.c	(revision 35093)
+++ ext/filter/sanitizing_filters.c	(revision 35401)
@@ -51,7 +51,9 @@
 	}
 
 	smart_str_0(&str);
-	efree(Z_STRVAL_P(value));
+	if (!IS_INTERNED(Z_STRVAL_P(value))) {
+		efree(Z_STRVAL_P(value));
+	}
 	Z_STRVAL_P(value) = str.c;
 	Z_STRLEN_P(value) = str.len;
 }
@@ -102,7 +104,9 @@
 		s++;	
 	}
 	*p = '\0';
-	efree(Z_STRVAL_P(value));
+	if (!IS_INTERNED(Z_STRVAL_P(value))) {
+		efree(Z_STRVAL_P(value));
+	}
 	Z_STRVAL_P(value) = (char *)str;
 	Z_STRLEN_P(value) = p - str;
 }
@@ -131,7 +135,9 @@
 	}
 	/* update zval string data */
 	buf[c] = '\0';
-	efree(Z_STRVAL_P(value));
+	if (!IS_INTERNED(Z_STRVAL_P(value))) {
+		efree(Z_STRVAL_P(value));
+	}
 	Z_STRVAL_P(value) = (char *)buf;
 	Z_STRLEN_P(value) = c;
 }
@@ -169,7 +175,9 @@
 	}
 	/* update zval string data */
 	buf[c] = '\0';
-	efree(Z_STRVAL_P(value));
+	if (!IS_INTERNED(Z_STRVAL_P(value))) {
+		efree(Z_STRVAL_P(value));
+	}
 	Z_STRVAL_P(value) = (char *)buf;
 	Z_STRLEN_P(value) = c;
 }
@@ -254,7 +262,9 @@
 		quotes = ENT_NOQUOTES;
 	}
 	buf = php_escape_html_entities_ex(Z_STRVAL_P(value), Z_STRLEN_P(value), &len, 1, quotes, SG(default_charset), 0 TSRMLS_CC);
-	efree(Z_STRVAL_P(value));
+	if (!IS_INTERNED(Z_STRVAL_P(value))) {
+		efree(Z_STRVAL_P(value));
+	}
 	Z_STRVAL_P(value) = buf;
 	Z_STRLEN_P(value) = len;
 }
@@ -345,7 +353,9 @@
 	/* just call php_addslashes quotes */
 	buf = php_addslashes(Z_STRVAL_P(value), Z_STRLEN_P(value), &len, 0 TSRMLS_CC);
 
-	efree(Z_STRVAL_P(value));
+	if (!IS_INTERNED(Z_STRVAL_P(value))) {
+		efree(Z_STRVAL_P(value));
+	}
 	Z_STRVAL_P(value) = buf;
 	Z_STRLEN_P(value) = len;
 }
Index: Zend/zend.c
===================================================================
--- Zend/zend.c	(revision 35093)
+++ Zend/zend.c	(revision 35401)
@@ -693,6 +693,7 @@
 	EG(user_exception_handler) = NULL;
 #endif
 
+	zend_interned_strings_init(TSRMLS_C);
 	zend_startup_builtin_functions(TSRMLS_C);
 	zend_register_standard_constants(TSRMLS_C);
 	zend_register_auto_global("GLOBALS", sizeof("GLOBALS") - 1, NULL TSRMLS_CC);
@@ -781,6 +782,8 @@
 	GLOBAL_CONSTANTS_TABLE = NULL;
 #endif
 	zend_destroy_rsrc_list_dtors();
+
+	zend_interned_strings_dtor(TSRMLS_C);
 }
 /* }}} */
 
Index: Zend/zend_hash.c
===================================================================
--- Zend/zend_hash.c	(revision 35093)
+++ Zend/zend_hash.c	(revision 35401)
@@ -20,6 +20,7 @@
 /* $Id: zend_hash.c 293155 2010-01-05 20:46:53Z sebastian $ */
 
 #include "zend.h"
+#include "zend_globals.h"
 
 #define CONNECT_TO_BUCKET_DLLIST(element, list_head)		\
 	(element)->pNext = (list_head);							\
@@ -210,8 +211,8 @@
 
 	p = ht->arBuckets[nIndex];
 	while (p != NULL) {
-		if ((p->h == h) && (p->nKeyLength == nKeyLength)) {
-			if (!memcmp(p->arKey, arKey, nKeyLength)) {
+		if (p->arKey == arKey ||
+			((p->h == h) && (p->nKeyLength == nKeyLength) && !memcmp(p->arKey, arKey, nKeyLength))) {
 				if (flag & HASH_ADD) {
 					return FAILURE;
 				}
@@ -232,16 +233,24 @@
 				}
 				HANDLE_UNBLOCK_INTERRUPTIONS();
 				return SUCCESS;
-			}
 		}
 		p = p->pNext;
 	}
 	
-	p = (Bucket *) pemalloc(sizeof(Bucket) - 1 + nKeyLength, ht->persistent);
-	if (!p) {
-		return FAILURE;
+	if (IS_INTERNED(arKey)) {
+		p = (Bucket *) pemalloc(sizeof(Bucket), ht->persistent);
+		if (!p) {
+			return FAILURE;
+		}
+		p->arKey = (char*)arKey;
+	} else {
+		p = (Bucket *) pemalloc(sizeof(Bucket) + nKeyLength, ht->persistent);
+		if (!p) {
+			return FAILURE;
+		}
+		p->arKey = (char*)(p + 1);
+		memcpy(p->arKey, arKey, nKeyLength);
 	}
-	memcpy(p->arKey, arKey, nKeyLength);
 	p->nKeyLength = nKeyLength;
 	INIT_DATA(ht, p, pData, nDataSize);
 	p->h = h;
@@ -276,8 +285,8 @@
 	
 	p = ht->arBuckets[nIndex];
 	while (p != NULL) {
-		if ((p->h == h) && (p->nKeyLength == nKeyLength)) {
-			if (!memcmp(p->arKey, arKey, nKeyLength)) {
+		if (p->arKey == arKey ||
+			((p->h == h) && (p->nKeyLength == nKeyLength) && !memcmp(p->arKey, arKey, nKeyLength))) {
 				if (flag & HASH_ADD) {
 					return FAILURE;
 				}
@@ -298,17 +307,25 @@
 				}
 				HANDLE_UNBLOCK_INTERRUPTIONS();
 				return SUCCESS;
-			}
 		}
 		p = p->pNext;
 	}
 	
-	p = (Bucket *) pemalloc(sizeof(Bucket) - 1 + nKeyLength, ht->persistent);
-	if (!p) {
-		return FAILURE;
+	if (IS_INTERNED(arKey)) {
+		p = (Bucket *) pemalloc(sizeof(Bucket), ht->persistent);
+		if (!p) {
+			return FAILURE;
+		}
+		p->arKey = (char*)arKey;
+	} else {
+		p = (Bucket *) pemalloc(sizeof(Bucket) + nKeyLength, ht->persistent);
+		if (!p) {
+			return FAILURE;
+		}
+		p->arKey = (char*)(p + 1);
+		memcpy(p->arKey, arKey, nKeyLength);
 	}
 
-	memcpy(p->arKey, arKey, nKeyLength);
 	p->nKeyLength = nKeyLength;
 	INIT_DATA(ht, p, pData, nDataSize);
 	p->h = h;
@@ -380,10 +397,11 @@
 		}
 		p = p->pNext;
 	}
-	p = (Bucket *) pemalloc_rel(sizeof(Bucket) - 1, ht->persistent);
+	p = (Bucket *) pemalloc_rel(sizeof(Bucket), ht->persistent);
 	if (!p) {
 		return FAILURE;
 	}
+	p->arKey = NULL;
 	p->nKeyLength = 0; /* Numeric indices are marked by making the nKeyLength == 0 */
 	p->h = h;
 	INIT_DATA(ht, p, pData, nDataSize);
@@ -877,11 +895,10 @@
 
 	p = ht->arBuckets[nIndex];
 	while (p != NULL) {
-		if ((p->h == h) && (p->nKeyLength == nKeyLength)) {
-			if (!memcmp(p->arKey, arKey, nKeyLength)) {
+		if (p->arKey == arKey ||
+			((p->h == h) && (p->nKeyLength == nKeyLength) && !memcmp(p->arKey, arKey, nKeyLength))) {
 				*pData = p->pData;
 				return SUCCESS;
-			}
 		}
 		p = p->pNext;
 	}
@@ -904,11 +921,10 @@
 
 	p = ht->arBuckets[nIndex];
 	while (p != NULL) {
-		if ((p->h == h) && (p->nKeyLength == nKeyLength)) {
-			if (!memcmp(p->arKey, arKey, nKeyLength)) {
+		if (p->arKey == arKey ||
+			((p->h == h) && (p->nKeyLength == nKeyLength) && !memcmp(p->arKey, arKey, nKeyLength))) {
 				*pData = p->pData;
 				return SUCCESS;
-			}
 		}
 		p = p->pNext;
 	}
@@ -929,10 +945,9 @@
 
 	p = ht->arBuckets[nIndex];
 	while (p != NULL) {
-		if ((p->h == h) && (p->nKeyLength == nKeyLength)) {
-			if (!memcmp(p->arKey, arKey, nKeyLength)) {
+		if (p->arKey == arKey ||
+			((p->h == h) && (p->nKeyLength == nKeyLength) && !memcmp(p->arKey, arKey, nKeyLength))) {
 				return 1;
-			}
 		}
 		p = p->pNext;
 	}
@@ -955,10 +970,9 @@
 
 	p = ht->arBuckets[nIndex];
 	while (p != NULL) {
-		if ((p->h == h) && (p->nKeyLength == nKeyLength)) {
-			if (!memcmp(p->arKey, arKey, nKeyLength)) {
+		if (p->arKey == arKey ||
+			((p->h == h) && (p->nKeyLength == nKeyLength) && !memcmp(p->arKey, arKey, nKeyLength))) {
 				return 1;
-			}
 		}
 		p = p->pNext;
 	}
@@ -1282,7 +1296,7 @@
 		}
 
 		if (p->nKeyLength != str_length) {
-			Bucket *q = (Bucket *) pemalloc(sizeof(Bucket) - 1 + str_length, ht->persistent);
+			Bucket *q = (Bucket *) pemalloc(sizeof(Bucket) + str_length, ht->persistent);
 
 			q->nKeyLength = str_length;
 			if (p->pData == &p->pDataPtr) {
@@ -1316,6 +1330,7 @@
 		if (key_type == HASH_KEY_IS_LONG) {
 			p->h = num_index;
 		} else {
+			p->arKey = (char*)(p+1);
 			memcpy(p->arKey, str_index, str_length);
 			p->h = zend_inline_hash_func(str_index, str_length);
 		}
Index: Zend/zend.h
===================================================================
--- Zend/zend.h	(revision 35093)
+++ Zend/zend.h	(revision 35401)
@@ -237,6 +237,7 @@
 #include "zend_alloc.h"
 
 #include "zend_types.h"
+#include "zend_string.h"
 
 #ifdef HAVE_LIMITS_H
 # include <limits.h>
@@ -599,8 +599,8 @@
 
 /* FIXME: Check if we can save if (ptr) too */
 
-#define STR_FREE(ptr) if (ptr) { efree(ptr); }
-#define STR_FREE_REL(ptr) if (ptr) { efree_rel(ptr); }
+#define STR_FREE(ptr) if (ptr && !IS_INTERNED(ptr)) { efree(ptr); }
+#define STR_FREE_REL(ptr) if (ptr && !IS_INTERNED(ptr)) { efree_rel(ptr); }
 
 #define STR_EMPTY_ALLOC() estrndup("", sizeof("")-1)
 
Index: Zend/zend_hash.h
===================================================================
--- Zend/zend_hash.h	(revision 35093)
+++ Zend/zend_hash.h	(revision 35401)
@@ -60,7 +60,7 @@
 	struct bucket *pListLast;
 	struct bucket *pNext;
 	struct bucket *pLast;
-	char arKey[1]; /* Must be last element */
+	char *arKey;
 } Bucket;
 
 typedef struct _hashtable {
Index: Zend/zend_execute.c
===================================================================
--- Zend/zend_execute.c	(revision 35093)
+++ Zend/zend_execute.c	(revision 35401)
@@ -611,12 +611,24 @@
 		}
 
 		if (T->str_offset.offset >= Z_STRLEN_P(T->str_offset.str)) {
-			Z_STRVAL_P(T->str_offset.str) = (char *) erealloc(Z_STRVAL_P(T->str_offset.str), T->str_offset.offset+1+1);
+			if (IS_INTERNED(Z_STRVAL_P(T->str_offset.str))) {
+				char *tmp = (char *) emalloc(T->str_offset.offset+1+1);
+
+				memcpy(tmp, Z_STRVAL_P(T->str_offset.str), T->str_offset.offset+1+1);
+				Z_STRVAL_P(T->str_offset.str) = tmp;
+			} else {
+				Z_STRVAL_P(T->str_offset.str) = (char *) erealloc(Z_STRVAL_P(T->str_offset.str), T->str_offset.offset+1+1);
+			}
 			memset(Z_STRVAL_P(T->str_offset.str) + Z_STRLEN_P(T->str_offset.str),
 			       ' ',
 			       T->str_offset.offset - Z_STRLEN_P(T->str_offset.str));
 			Z_STRVAL_P(T->str_offset.str)[T->str_offset.offset+1] = 0;
 			Z_STRLEN_P(T->str_offset.str) = T->str_offset.offset+1;
+		} else if (IS_INTERNED(Z_STRVAL_P(T->str_offset.str))) {
+			char *tmp = (char *) emalloc(Z_STRLEN_P(T->str_offset.str) + 1);
+
+			memcpy(tmp, Z_STRVAL_P(T->str_offset.str), Z_STRLEN_P(T->str_offset.str) + 1);
+			Z_STRVAL_P(T->str_offset.str) = tmp;
 		}
 
 		if (Z_TYPE_P(value) != IS_STRING) {
@@ -804,7 +816,11 @@
 				hval = Z_HASH_P(dim);
 			} else {
 				ZEND_HANDLE_NUMERIC_EX(offset_key, offset_key_length+1, index, goto num_index);
-				hval = zend_hash_func(offset_key, offset_key_length+1);
+				if (IS_INTERNED(offset_key)) {
+					hval = INTERNED_HASH(offset_key);
+				} else {
+					hval = zend_hash_func(offset_key, offset_key_length+1);
+				}
 			}	
 fetch_string_dim:
 			if (zend_hash_quick_find(ht, offset_key, offset_key_length+1, hval, (void **) &retval) == FAILURE) {
Index: Zend/zend_variables.c
===================================================================
--- Zend/zend_variables.c	(revision 35093)
+++ Zend/zend_variables.c	(revision 35401)
@@ -77,7 +77,9 @@
 		case IS_STRING:
 		case IS_CONSTANT:
 			CHECK_ZVAL_STRING_REL(zvalue);
-			free(zvalue->value.str.val);
+			if (!IS_INTERNED(zvalue->value.str.val)) {
+				free(zvalue->value.str.val);
+			}
 			break;
 		case IS_ARRAY:
 		case IS_CONSTANT_ARRAY:
@@ -117,7 +119,9 @@
 		case IS_CONSTANT:
 		case IS_STRING:
 			CHECK_ZVAL_STRING_REL(zvalue);
-			zvalue->value.str.val = (char *) estrndup_rel(zvalue->value.str.val, zvalue->value.str.len);
+			if (!IS_INTERNED(zvalue->value.str.val)) {
+				zvalue->value.str.val = (char *) estrndup_rel(zvalue->value.str.val, zvalue->value.str.len);
+			}
 			break;
 		case IS_ARRAY:
 		case IS_CONSTANT_ARRAY: {
Index: Zend/zend_builtin_functions.c
===================================================================
--- Zend/zend_builtin_functions.c	(revision 35093)
+++ Zend/zend_builtin_functions.c	(revision 35401)
@@ -681,7 +681,7 @@
 		zval_ptr_dtor(&val_free);
 	}
 	c.flags = case_sensitive; /* non persistent */
-	c.name = zend_strndup(name, name_len);
+	c.name = IS_INTERNED(name) ? name : zend_strndup(name, name_len);
 	c.name_len = name_len+1;
 	c.module_number = PHP_USER_CONSTANT;
 	if (zend_register_constant(&c TSRMLS_CC) == SUCCESS) {
Index: Zend/zend_language_scanner.l
===================================================================
--- Zend/zend_language_scanner.l	(revision 35093)
+++ Zend/zend_language_scanner.l	(revision 35401)
@@ -408,7 +408,13 @@
 ZEND_API int zend_prepare_string_for_scanning(zval *str, char *filename TSRMLS_DC)
 {
 	/* enforce two trailing NULLs for flex... */
-	str->value.str.val = safe_erealloc(str->value.str.val, 1, str->value.str.len, ZEND_MMAP_AHEAD);
+	if (IS_INTERNED(str->value.str.val)) {
+		char *tmp = safe_emalloc(1, str->value.str.len, ZEND_MMAP_AHEAD);
+		memcpy(tmp, str->value.str.val, str->value.str.len + ZEND_MMAP_AHEAD);
+		str->value.str.val = tmp;
+	} else {
+		str->value.str.val = safe_erealloc(str->value.str.val, 1, str->value.str.len, ZEND_MMAP_AHEAD);
+	}
 
 	memset(str->value.str.val + str->value.str.len, 0, ZEND_MMAP_AHEAD);
 
Index: Zend/zend_constants.c
===================================================================
--- Zend/zend_constants.c	(revision 35093)
+++ Zend/zend_constants.c	(revision 35401)
@@ -32,13 +32,17 @@
 	if (!(c->flags & CONST_PERSISTENT)) {
 		zval_dtor(&c->value);
 	}
-	free(c->name);
+	if (!IS_INTERNED(c->name)) {
+		free(c->name);
+	}
 }
 
 
 void copy_zend_constant(zend_constant *c)
 {
-	c->name = zend_strndup(c->name, c->name_len - 1);
+	if (!IS_INTERNED(c->name)) {
+		c->name = zend_strndup(c->name, c->name_len - 1);
+	}
 	if (!(c->flags & CONST_PERSISTENT)) {
 		zval_copy_ctor(&c->value);
 	}
@@ -422,12 +436,14 @@
 		/* keep in mind that c->name_len already contains the '\0' */
 		lowercase_name = estrndup(c->name, c->name_len-1);
 		zend_str_tolower(lowercase_name, c->name_len-1);
+		lowercase_name = CG(new_interned_string)(lowercase_name, c->name_len, 1 TSRMLS_CC);
 		name = lowercase_name;
 	} else {
 		char *slash = strrchr(c->name, '\\');
 		if(slash) {
 			lowercase_name = estrndup(c->name, c->name_len-1);
 			zend_str_tolower(lowercase_name, slash-c->name);
+			lowercase_name = CG(new_interned_string)(lowercase_name, c->name_len, 1 TSRMLS_CC);
 			name = lowercase_name;
 		} else {
 			name = c->name;
@@ -437,13 +453,15 @@
 	if ((strncmp(name, "__COMPILER_HALT_OFFSET__", sizeof("__COMPILER_HALT_OFFSET__") - 1) == 0) ||
 			zend_hash_add(EG(zend_constants), name, c->name_len, (void *) c, sizeof(zend_constant), NULL)==FAILURE) {
 		zend_error(E_NOTICE,"Constant %s already defined", name);
-		free(c->name);
+		if (!IS_INTERNED(c->name)) {
+			free(c->name);
+		}
 		if (!(c->flags & CONST_PERSISTENT)) {
 			zval_dtor(&c->value);
 		}
 		ret = FAILURE;
 	}
-	if (lowercase_name) {
+	if (lowercase_name && !IS_INTERNED(lowercase_name)) {
 		efree(lowercase_name);
 	}
 	return ret;
Index: Zend/zend_compile.c
===================================================================
--- Zend/zend_compile.c	(revision 35093)
+++ Zend/zend_compile.c	(revision 35401)
@@ -73,7 +73,9 @@
 
 static void zend_duplicate_property_info(zend_property_info *property_info) /* {{{ */
 {
-	property_info->name = estrndup(property_info->name, property_info->name_length);
+	if (!IS_INTERNED(property_info->name)) {		
+		property_info->name = estrndup(property_info->name, property_info->name_length);
+	}
 	if (property_info->doc_comment) {
 		property_info->doc_comment = estrndup(property_info->doc_comment, property_info->doc_comment_len);
 	}
@@ -83,14 +85,18 @@
 
 static void zend_duplicate_property_info_internal(zend_property_info *property_info) /* {{{ */
 {
-	property_info->name = zend_strndup(property_info->name, property_info->name_length);
+	if (!IS_INTERNED(property_info->name)) {		
+		property_info->name = zend_strndup(property_info->name, property_info->name_length);
+	}
 }
 /* }}} */
 
 
 static void zend_destroy_property_info(zend_property_info *property_info) /* {{{ */
 {
-	efree(property_info->name);
+    if (!IS_INTERNED(property_info->name)) {
+		efree(property_info->name);
+	}
 	if (property_info->doc_comment) {
 		efree(property_info->doc_comment);
 	}
@@ -100,7 +106,9 @@
 
 static void zend_destroy_property_info_internal(zend_property_info *property_info) /* {{{ */
 {
-	free(property_info->name);
+    if (!IS_INTERNED(property_info->name)) {
+		free(property_info->name);
+	}
 }
 /* }}} */
 
@@ -292,17 +300,20 @@
 }
 /* }}} */
 
-static int lookup_cv(zend_op_array *op_array, char* name, int name_len) /* {{{ */
+static int lookup_cv(zend_op_array *op_array, char* name, int name_len TSRMLS_DC) /* {{{ */
 {
 	int i = 0;
 	ulong hash_value = zend_inline_hash_func(name, name_len+1);
 
 	while (i < op_array->last_var) {
-		if (op_array->vars[i].hash_value == hash_value &&
-		    op_array->vars[i].name_len == name_len &&
-		    strcmp(op_array->vars[i].name, name) == 0) {
-		  efree(name);
-		  return i;
+		if (op_array->vars[i].name == name ||
+		    (op_array->vars[i].hash_value == hash_value &&
+		     op_array->vars[i].name_len == name_len &&
+		     memcmp(op_array->vars[i].name, name, name_len) == 0)) {
+			if (!IS_INTERNED(name)) {			
+				efree(name);
+			}
+			return i;
 		}
 		i++;
 	}
@@ -312,7 +323,7 @@
 		op_array->size_var += 16; /* FIXME */
 		op_array->vars = erealloc(op_array->vars, op_array->size_var*sizeof(zend_compiled_variable));
 	}
-	op_array->vars[i].name = name; /* estrndup(name, name_len); */
+	op_array->vars[i].name = CG(new_interned_string)(name, name_len + 1, 1 TSRMLS_CC);
 	op_array->vars[i].name_len = name_len;
 	op_array->vars[i].hash_value = hash_value;
 	return i;
@@ -338,6 +349,13 @@
 		op_array->size_literal += 16; /* FIXME */
 		op_array->literals = (zend_literal*)erealloc(op_array->literals, op_array->size_literal * sizeof(zend_literal));
 	}
+	if (Z_TYPE_P(zv) == IS_STRING || Z_TYPE_P(zv) == IS_CONSTANT) {
+		zval *z = (zval*)zv;
+		TSRMLS_FETCH();
+
+		Z_STRVAL_P(z) =
+			CG(new_interned_string)(Z_STRVAL_P(zv), Z_STRLEN_P(zv) + 1, 1 TSRMLS_CC);
+	}
 	CONSTANT_EX(op_array, i) = *zv;
 	Z_SET_REFCOUNT(CONSTANT_EX(op_array, i), 2);
 	Z_SET_ISREF(CONSTANT_EX(op_array, i));
@@ -463,7 +481,8 @@
 		    (CG(active_op_array)->last == 0 ||
 		     CG(active_op_array)->opcodes[CG(active_op_array)->last-1].opcode != ZEND_BEGIN_SILENCE)) {
 			result->op_type = IS_CV;
-			result->u.op.var = lookup_cv(CG(active_op_array), varname->u.constant.value.str.val, varname->u.constant.value.str.len);
+			result->u.op.var = lookup_cv(CG(active_op_array), varname->u.constant.value.str.val, varname->u.constant.value.str.len TSRMLS_CC);
+			varname->u.constant.value.str.val = CG(active_op_array)->vars[result->u.op.var].name;
 			result->EA = 0;
 			return;
 		}
@@ -1078,7 +1097,7 @@
 
 				this_var = opline_ptr->result.var;
 				if (CG(active_op_array)->this_var == -1) {
-					CG(active_op_array)->this_var = lookup_cv(CG(active_op_array), Z_STRVAL(CONSTANT(opline_ptr->op1.constant)), Z_STRLEN(CONSTANT(opline_ptr->op1.constant)));
+					CG(active_op_array)->this_var = lookup_cv(CG(active_op_array), Z_STRVAL(CONSTANT(opline_ptr->op1.constant)), Z_STRLEN(CONSTANT(opline_ptr->op1.constant)) TSRMLS_CC);
 					Z_TYPE(CONSTANT(opline_ptr->op1.constant)) = IS_NULL;
 				} else {
 					zend_del_literal(CG(active_op_array), opline_ptr->op1.constant);
@@ -1090,7 +1109,7 @@
 					variable->u.var = CG(active_op_array)->this_var;
 				}
 		    } else if (CG(active_op_array)->this_var == -1) {
-				CG(active_op_array)->this_var = lookup_cv(CG(active_op_array), estrndup("this", sizeof("this")-1), sizeof("this")-1);
+				CG(active_op_array)->this_var = lookup_cv(CG(active_op_array), estrndup("this", sizeof("this")-1), sizeof("this")-1 TSRMLS_CC);
 			}
 		}
 
@@ -1290,7 +1309,6 @@
 	}
 
 	function_token->u.op_array = CG(active_op_array);
-	lcname = zend_str_tolower_dup(name, name_len);
 
 	orig_interactive = CG(interactive);
 	CG(interactive) = 0;
@@ -1308,6 +1326,8 @@
 	op_array.line_start = zend_get_compiled_lineno(TSRMLS_C);
 
 	if (is_method) {
+		lcname = CG(new_interned_string)(zend_str_tolower_dup(name, name_len), name_len + 1, 1 TSRMLS_CC);
+
 		if (zend_hash_add(&CG(active_class_entry)->function_table, lcname, name_len+1, &op_array, sizeof(zend_op_array), (void **) &CG(active_op_array)) == FAILURE) {
 			zend_error(E_COMPILE_ERROR, "Cannot redeclare %s::%s()", CG(active_class_entry)->name, name);
 		}
@@ -1422,7 +1442,9 @@
 			free_alloca(short_class_lcname, use_heap);
 		}
 
-		efree(lcname);
+		if (!IS_INTERNED(lcname)) {
+			efree(lcname);
+		}
 	} else {
 		zend_op *opline = get_next_op(CG(active_op_array) TSRMLS_CC);
 		zval key;
@@ -1435,9 +1457,10 @@
 			zval_copy_ctor(&tmp.u.constant);
 			zend_do_build_namespace_name(&tmp, &tmp, function_name TSRMLS_CC);
 			op_array.function_name = Z_STRVAL(tmp.u.constant);
-			efree(lcname);
 			name_len = Z_STRLEN(tmp.u.constant);
 			lcname = zend_str_tolower_dup(Z_STRVAL(tmp.u.constant), name_len);
+		} else {
+			lcname = zend_str_tolower_dup(name, name_len);
 		}
 
 		opline->opcode = ZEND_DECLARE_FUNCTION;
@@ -1602,7 +1625,7 @@
 	}
 	CG(active_op_array)->arg_info = erealloc(CG(active_op_array)->arg_info, sizeof(zend_arg_info)*(CG(active_op_array)->num_args));
 	cur_arg_info = &CG(active_op_array)->arg_info[CG(active_op_array)->num_args-1];
-	cur_arg_info->name = estrndup(varname->u.constant.value.str.val, varname->u.constant.value.str.len);
+	cur_arg_info->name = CG(new_interned_string)(estrndup(varname->u.constant.value.str.val, varname->u.constant.value.str.len), varname->u.constant.value.str.len + 1, 1 TSRMLS_CC);
 	cur_arg_info->name_len = varname->u.constant.value.str.len;
 	cur_arg_info->array_type_hint = 0;
 	cur_arg_info->allow_null = 1;
@@ -1616,6 +1639,7 @@
 			if (ZEND_FETCH_CLASS_DEFAULT == zend_get_class_fetch_type(Z_STRVAL(class_type->u.constant), Z_STRLEN(class_type->u.constant))) {
 				zend_resolve_class_name(class_type, &opline->extended_value, 1 TSRMLS_CC);
 			}
+			class_type->u.constant.value.str.val = CG(new_interned_string)(class_type->u.constant.value.str.val, class_type->u.constant.value.str.len + 1, 1 TSRMLS_CC);
 			cur_arg_info->class_name = class_type->u.constant.value.str.val;
 			cur_arg_info->class_name_len = class_type->u.constant.value.str.len;
 			if (op == ZEND_RECV_INIT) {
@@ -1689,7 +1713,7 @@
 	lc_name = zend_str_tolower_dup(Z_STRVAL(CONSTANT(literal)), Z_STRLEN(CONSTANT(literal)));
 	ZVAL_STRINGL(&c, lc_name, Z_STRLEN(CONSTANT(literal)), 0);
 	lc_literal = zend_add_literal(CG(active_op_array), &c);
-	Z_HASH_P(&CONSTANT(lc_literal)) = zend_hash_func(lc_name, Z_STRLEN(CONSTANT(literal))+1);
+	Z_HASH_P(&CONSTANT(lc_literal)) = zend_hash_func(Z_STRVAL(c), Z_STRLEN(c)+1);
 if (literal + 1 != lc_literal) {
 zend_error(E_ERROR, "Internal error 2 ???");
 }
@@ -1716,7 +1740,7 @@
 	}
 	ZVAL_STRINGL(&c, lc_name, lc_len, 0);
 	lc_literal = zend_add_literal(CG(active_op_array), &c);
-	Z_HASH_P(&CONSTANT(lc_literal)) = zend_hash_func(lc_name, lc_len+1);
+	Z_HASH_P(&CONSTANT(lc_literal)) = zend_hash_func(Z_STRVAL(c), Z_STRLEN(c)+1);
 if (literal + 1 != lc_literal) {
 zend_error(E_ERROR, "Internal error 4 ???");
 }
@@ -2482,7 +2506,7 @@
 }
 /* }}} */
 
-void zend_do_begin_catch(znode *try_token, znode *class_name, const znode *catch_var, znode *first_catch TSRMLS_DC) /* {{{ */
+void zend_do_begin_catch(znode *try_token, znode *class_name, znode *catch_var, znode *first_catch TSRMLS_DC) /* {{{ */
 {
 	long catch_op_number;
 	zend_op *opline;
@@ -2506,7 +2530,8 @@
 	opline->opcode = ZEND_CATCH;
 	SET_NODE(opline->op1, &catch_class);
 	opline->op2_type = IS_CV;
-	opline->op2.var = lookup_cv(CG(active_op_array), catch_var->u.constant.value.str.val, catch_var->u.constant.value.str.len);
+	opline->op2.var = lookup_cv(CG(active_op_array), catch_var->u.constant.value.str.val, catch_var->u.constant.value.str.len TSRMLS_CC);
+	catch_var->u.constant.value.str.val = CG(active_op_array)->vars[opline->op2.var].name;
 	opline->result.num = 0; /* 1 means it's the last catch in the block */
 
 	try_token->u.opline_num = catch_op_number;
@@ -3812,7 +3837,7 @@
 		CG(doc_comment_len) = 0;
 	}
 
-	zend_declare_property_ex(CG(active_class_entry), var_name->u.constant.value.str.val, var_name->u.constant.value.str.len, property, access_type, comment, comment_len TSRMLS_CC);
+	zend_declare_property_ex(CG(active_class_entry), CG(new_interned_string)(var_name->u.constant.value.str.val, var_name->u.constant.value.str.len + 1, 0 TSRMLS_CC), var_name->u.constant.value.str.len, property, access_type, comment, comment_len TSRMLS_CC);
 	efree(var_name->u.constant.value.str.val);
 }
 /* }}} */
@@ -3828,7 +3853,7 @@
 	ALLOC_ZVAL(property);
 	*property = value->u.constant;
 
-	if (zend_hash_add(&CG(active_class_entry)->constants_table, var_name->u.constant.value.str.val, var_name->u.constant.value.str.len+1, &property, sizeof(zval *), NULL)==FAILURE) {
+	if (zend_hash_add(&CG(active_class_entry)->constants_table, CG(new_interned_string)(var_name->u.constant.value.str.val, var_name->u.constant.value.str.len+1, 0 TSRMLS_CC), var_name->u.constant.value.str.len+1, &property, sizeof(zval *), NULL)==FAILURE) {
 		FREE_ZVAL(property);
 		zend_error(E_COMPILE_ERROR, "Cannot redefine class constant %s::%s", CG(active_class_entry)->name, var_name->u.constant.value.str.val);
 	}
Index: Zend/zend_compile.h
===================================================================
--- Zend/zend_compile.h	(revision 35093)
+++ Zend/zend_compile.h	(revision 35401)
@@ -463,7 +463,7 @@
 void zend_do_fetch_lexical_variable(znode *varname, zend_bool is_ref TSRMLS_DC);
 
 void zend_do_try(znode *try_token TSRMLS_DC);
-void zend_do_begin_catch(znode *try_token, znode *catch_class, const znode *catch_var, znode *first_catch TSRMLS_DC);
+void zend_do_begin_catch(znode *try_token, znode *catch_class, znode *catch_var, znode *first_catch TSRMLS_DC);
 void zend_do_end_catch(const znode *try_token TSRMLS_DC);
 void zend_do_throw(const znode *expr TSRMLS_DC);
 
Index: Zend/zend_API.c
===================================================================
--- Zend/zend_API.c	(revision 35093)
+++ Zend/zend_API.c	(revision 35401)
@@ -1895,9 +1895,11 @@
 		}
 		fname_len = strlen(ptr->fname);
- 		lowercase_name = zend_str_tolower_dup(ptr->fname, fname_len);
+		lowercase_name = CG(new_interned_string)(zend_str_tolower_dup(ptr->fname, fname_len), fname_len + 1, 1 TSRMLS_CC);
 		if (zend_hash_add(target_function_table, lowercase_name, fname_len+1, &function, sizeof(zend_function), (void**)&reg_function) == FAILURE) {
 			unload=1;
-			efree(lowercase_name);
+			if (!IS_INTERNED(lowercase_name)) {
+				efree(lowercase_name);
+			}
 			break;
 		}
 		if (scope) {
@@ -1939,7 +1942,9 @@
 		}
 		ptr++;
 		count++;
-		efree(lowercase_name);
+		if (!IS_INTERNED(lowercase_name)) {
+			efree(lowercase_name);
+		}
 	}
 	if (unload) { /* before unloading, display all remaining bad function in the module */
 		if (scope) {
@@ -2169,7 +2174,7 @@
 static zend_class_entry *do_register_internal_class(zend_class_entry *orig_class_entry, zend_uint ce_flags TSRMLS_DC) /* {{{ */
 {
 	zend_class_entry *class_entry = malloc(sizeof(zend_class_entry));
-	char *lowercase_name = malloc(orig_class_entry->name_length + 1);
+	char *lowercase_name = emalloc(orig_class_entry->name_length + 1);
 	*class_entry = *orig_class_entry;
 
 	class_entry->type = ZEND_INTERNAL_CLASS;
@@ -2182,8 +2187,11 @@
 	}
 
 	zend_str_tolower_copy(lowercase_name, orig_class_entry->name, class_entry->name_length);
+	lowercase_name = CG(new_interned_string)(lowercase_name, class_entry->name_length + 1, 1 TSRMLS_CC);
 	zend_hash_update(CG(class_table), lowercase_name, class_entry->name_length+1, &class_entry, sizeof(zend_class_entry *), NULL);
-	free(lowercase_name);
+	if (!IS_INTERNED(lowercase_name)) {
+		efree(lowercase_name);
+	}
 	return class_entry;
 }
 /* }}} */
@@ -3071,6 +3079,7 @@
 {
 	zend_property_info property_info;
 	HashTable *target_symbol_table;
+	char *interned_name;
 
 	if (!(access_type & ZEND_ACC_PPP_MASK)) {
 		access_type |= ZEND_ACC_PUBLIC;
@@ -3098,7 +3107,6 @@
 				int priv_name_length;
 
 				zend_mangle_property_name(&priv_name, &priv_name_length, ce->name, ce->name_length, name, name_length, ce->type & ZEND_INTERNAL_CLASS);
-				zend_hash_update(target_symbol_table, priv_name, priv_name_length+1, &property, sizeof(zval *), NULL);
 				property_info.name = priv_name;
 				property_info.name_length = priv_name_length;
 			}
@@ -3108,7 +3116,6 @@
 				int prot_name_length;
 
 				zend_mangle_property_name(&prot_name, &prot_name_length, "*", 1, name, name_length, ce->type & ZEND_INTERNAL_CLASS);
-				zend_hash_update(target_symbol_table, prot_name, prot_name_length+1, &property, sizeof(zval *), NULL);
 				property_info.name = prot_name;
 				property_info.name_length = prot_name_length;
 			}
@@ -3122,11 +3129,27 @@
 				zend_hash_del(target_symbol_table, prot_name, prot_name_length+1);
 				pefree(prot_name, ce->type & ZEND_INTERNAL_CLASS);
 			}
-			zend_hash_update(target_symbol_table, name, name_length+1, &property, sizeof(zval *), NULL);
-			property_info.name = ce->type & ZEND_INTERNAL_CLASS ? zend_strndup(name, name_length) : estrndup(name, name_length);
+			if (IS_INTERNED(name)) {
+				property_info.name = (char*)name;
+			} else {
+				property_info.name = ce->type & ZEND_INTERNAL_CLASS ? zend_strndup(name, name_length) : estrndup(name, name_length);
+			}
 			property_info.name_length = name_length;
 			break;
 	}
+
+	interned_name = CG(new_interned_string)(property_info.name, property_info.name_length+1, 0 TSRMLS_CC);
+	if (interned_name != property_info.name) {
+		if (ce->type == ZEND_USER_CLASS) {
+			efree(property_info.name);
+		} else {
+			free(property_info.name);
+		}
+		property_info.name = interned_name;
+	}
+
+	zend_hash_update(target_symbol_table, property_info.name, property_info.name_length+1, &property, sizeof(zval *), NULL);
+
 	property_info.flags = access_type;
 	property_info.h = zend_get_hash_value(property_info.name, property_info.name_length+1);
 
Index: Zend/zend_operators.c
===================================================================
--- Zend/zend_operators.c	(revision 35093)
+++ Zend/zend_operators.c	(revision 35401)
@@ -1177,11 +1177,18 @@
 /* must support result==op1 */
 ZEND_API int add_char_to_string(zval *result, const zval *op1, const zval *op2) /* {{{ */
 {
-	Z_STRLEN_P(result) = Z_STRLEN_P(op1) + 1;
-	Z_STRVAL_P(result) = (char *) erealloc(Z_STRVAL_P(op1), Z_STRLEN_P(result)+1);
-	Z_STRVAL_P(result)[Z_STRLEN_P(result) - 1] = (char) Z_LVAL_P(op2);
-	Z_STRVAL_P(result)[Z_STRLEN_P(result)] = 0;
-	Z_TYPE_P(result) = IS_STRING;
+	int length = Z_STRLEN_P(op1) + 1;
+	char *buf;
+
+	if (IS_INTERNED(Z_STRVAL_P(op1))) {
+		buf = (char *) emalloc(length + 1);
+		memcpy(buf, Z_STRVAL_P(op1), Z_STRLEN_P(op1));
+	} else {
+		buf = (char *) erealloc(Z_STRVAL_P(op1), length + 1);
+	}
+	buf[length - 1] = (char) Z_LVAL_P(op2);
+	buf[length] = 0;
+	ZVAL_STRINGL(result, buf, length, 0);
 	return SUCCESS;
 }
 /* }}} */
@@ -1190,12 +1197,17 @@
 ZEND_API int add_string_to_string(zval *result, const zval *op1, const zval *op2) /* {{{ */
 {
 	int length = Z_STRLEN_P(op1) + Z_STRLEN_P(op2);
+	char *buf;
 
-	Z_STRVAL_P(result) = (char *) erealloc(Z_STRVAL_P(op1), length+1);
-	memcpy(Z_STRVAL_P(result)+Z_STRLEN_P(op1), Z_STRVAL_P(op2), Z_STRLEN_P(op2));
-	Z_STRVAL_P(result)[length] = 0;
-	Z_STRLEN_P(result) = length;
-	Z_TYPE_P(result) = IS_STRING;
+	if (IS_INTERNED(Z_STRVAL_P(op1))) {
+		buf = (char *) emalloc(length+1);
+		memcpy(buf, Z_STRVAL_P(op1), Z_STRLEN_P(op1));
+	} else {
+		buf = (char *) erealloc(Z_STRVAL_P(op1), length+1);
+	}
+	memcpy(buf + Z_STRLEN_P(op1), Z_STRVAL_P(op2), Z_STRLEN_P(op2));
+	buf[length] = 0;
+	ZVAL_STRINGL(result, buf, length, 0);
 	return SUCCESS;
 }
 /* }}} */
@@ -1224,7 +1236,7 @@
 	if (use_copy2) {
 		op2 = &op2_copy;
 	}
-	if (result==op1) {	/* special case, perform operations on result */
+	if (result==op1 && !IS_INTERNED(Z_STRVAL_P(op1))) {	/* special case, perform operations on result */
 		uint res_len = Z_STRLEN_P(op1) + Z_STRLEN_P(op2);
 
 		if (Z_STRLEN_P(result) < 0 || (int) (Z_STRLEN_P(op1) + Z_STRLEN_P(op2)) < 0) {
@@ -1239,12 +1251,13 @@
 		Z_STRVAL_P(result)[res_len]=0;
 		Z_STRLEN_P(result) = res_len;
 	} else {
-		Z_STRLEN_P(result) = Z_STRLEN_P(op1) + Z_STRLEN_P(op2);
-		Z_STRVAL_P(result) = (char *) emalloc(Z_STRLEN_P(result) + 1);
-		memcpy(Z_STRVAL_P(result), Z_STRVAL_P(op1), Z_STRLEN_P(op1));
-		memcpy(Z_STRVAL_P(result)+Z_STRLEN_P(op1), Z_STRVAL_P(op2), Z_STRLEN_P(op2));
-		Z_STRVAL_P(result)[Z_STRLEN_P(result)] = 0;
-		Z_TYPE_P(result) = IS_STRING;
+		int length = Z_STRLEN_P(op1) + Z_STRLEN_P(op2);
+		char *buf = (char *) emalloc(length + 1);
+
+		memcpy(buf, Z_STRVAL_P(op1), Z_STRLEN_P(op1));
+		memcpy(buf + Z_STRLEN_P(op1), Z_STRVAL_P(op2), Z_STRLEN_P(op2));
+		buf[length] = 0;
+		ZVAL_STRINGL(result, buf, length, 0);
 	}
 	if (use_copy1) {
 		zval_dtor(op1);
@@ -1668,6 +1681,12 @@
 		return;
 	}
 
+	if (IS_INTERNED(s)) {
+		s = (char*) emalloc(Z_STRLEN_P(str) + 1);
+		memcpy(s, Z_STRVAL_P(str), Z_STRLEN_P(str) + 1);
+		Z_STRVAL_P(str) = s;
+	}
+
 	while (pos >= 0) {
 		ch = s[pos];
 		if (ch >= 'a' && ch <= 'z') {
@@ -1753,7 +1772,9 @@
 
 				switch (is_numeric_string(Z_STRVAL_P(op1), Z_STRLEN_P(op1), &lval, &dval, 0)) {
 					case IS_LONG:
-						efree(Z_STRVAL_P(op1));
+						if (!IS_INTERNED(Z_STRVAL_P(op1))) {
+							efree(Z_STRVAL_P(op1));
+						}
 						if (lval == LONG_MAX) {
 							/* switch to double */
 							double d = (double)lval;
@@ -1763,7 +1784,9 @@
 						}
 						break;
 					case IS_DOUBLE:
-						efree(Z_STRVAL_P(op1));
+						if (!IS_INTERNED(Z_STRVAL_P(op1))) {
+							efree(Z_STRVAL_P(op1));
+						}
 						ZVAL_DOUBLE(op1, dval+1);
 						break;
 					default:
@@ -1879,6 +1902,9 @@
 {
 	int retval;
 
+	if (s1 == s2) {
+		return 0;
+	}
 	retval = memcmp(s1, s2, MIN(len1, len2));
 	if (!retval) {
 		return (len1 - len2);
@@ -1892,6 +1918,9 @@
 {
 	int retval;
 
+	if (s1 == s2) {
+		return 0;
+	}
 	retval = memcmp(s1, s2, MIN(length, MIN(len1, len2)));
 	if (!retval) {
 		return (MIN(length, len1) - MIN(length, len2));
@@ -1906,8 +1935,11 @@
 	int len;
 	int c1, c2;
 
+	if (s1 == s2) {
+		return 0;
+	}
+
 	len = MIN(len1, len2);
-
 	while (len--) {
 		c1 = zend_tolower((int)*(unsigned char *)s1++);
 		c2 = zend_tolower((int)*(unsigned char *)s2++);
@@ -1925,8 +1957,10 @@
 	int len;
 	int c1, c2;
 
+	if (s1 == s2) {
+		return 0;
+	}
 	len = MIN(length, MIN(len1, len2));
-
 	while (len--) {
 		c1 = zend_tolower((int)*(unsigned char *)s1++);
 		c2 = zend_tolower((int)*(unsigned char *)s2++);
Index: Zend/zend_globals.h
===================================================================
--- Zend/zend_globals.h	(revision 35093)
+++ Zend/zend_globals.h	(revision 35401)
@@ -140,6 +140,18 @@
 	HashTable *labels;
 	zend_stack labels_stack;
 
+	/* interned strings */
+	char *interned_strings_start;
+	char *interned_strings_end;
+	char *interned_strings_top;
+	char *interned_strings_snapshot_top;
+
+	HashTable interned_strings;
+
+	char *(*new_interned_string)(char *str, int len, int free_src TSRMLS_DC);
+	void (*interned_strings_snapshot)(TSRMLS_D);
+	void (*interned_strings_restore)(TSRMLS_D);
+
 #ifdef ZEND_MULTIBYTE
 	zend_encoding **script_encoding_list;
 	size_t script_encoding_list_size;
Index: Zend/zend_execute_API.c
===================================================================
--- Zend/zend_execute_API.c	(revision 35093)
+++ Zend/zend_execute_API.c	(revision 35401)
@@ -521,7 +521,9 @@
 				Z_STRLEN_P(p) -= ((colon - Z_STRVAL_P(p)) + 1);
 				if (inline_change) {
 					colon = estrndup(colon, Z_STRLEN_P(p));
-					efree(Z_STRVAL_P(p));
+					if (!IS_INTERNED(Z_STRVAL_P(p))) {
+						efree(Z_STRVAL_P(p));
+					}
 					Z_STRVAL_P(p) = colon;
 				} else {
 					Z_STRVAL_P(p) = colon + 1;
@@ -557,12 +559,12 @@
 					if (fix_save) {
 						save--;
 					}
-					if (inline_change) {
+					if (inline_change && !IS_INTERNED(save)) {
 						efree(save);
 					}
 					save = NULL;
 				}
-				if (inline_change && save && save != actual) {
+				if (inline_change && save && save != actual && !IS_INTERNED(save)) {
 					efree(save);
 				}
 				zend_error(E_NOTICE, "Use of undefined constant %s - assumed '%s'",  actual,  actual);
Index: Zend/zend_vm_def.h
===================================================================
--- Zend/zend_vm_def.h	(revision 35093)
+++ Zend/zend_vm_def.h	(revision 35401)
@@ -951,6 +951,8 @@
 */
 		if (OP1_TYPE == IS_CONST) {
 			hash_value = Z_HASH_P(varname);
+		} else if (IS_INTERNED(Z_STRVAL_P(varname))) {
+			hash_value = INTERNED_HASH(Z_STRVAL_P(varname));
 		} else {
 			hash_value = zend_hash_func(Z_STRVAL_P(varname), Z_STRLEN_P(varname)+1);
 		}
@@ -3115,7 +3117,11 @@
 					hval = Z_HASH_P(offset);
 				} else {
 					ZEND_HANDLE_NUMERIC_EX(Z_STRVAL_P(offset), Z_STRLEN_P(offset)+1, index, ZEND_VM_C_GOTO(num_index));
-					hval = zend_hash_func(Z_STRVAL_P(offset), Z_STRLEN_P(offset)+1);
+					if (IS_INTERNED(Z_STRVAL_P(offset))) {
+						hval = INTERNED_HASH(Z_STRVAL_P(offset));
+					} else {
+						hval = zend_hash_func(Z_STRVAL_P(offset), Z_STRLEN_P(offset)+1);
+					}
 				}
 				zend_hash_quick_update(Z_ARRVAL_P(array_ptr), Z_STRVAL_P(offset), Z_STRLEN_P(offset)+1, hval, &expr_ptr, sizeof(zval *), NULL);
 				break;
@@ -3459,7 +3465,11 @@
 							hval = Z_HASH_P(offset);
 						} else {
 							ZEND_HANDLE_NUMERIC_EX(Z_STRVAL_P(offset), Z_STRLEN_P(offset)+1, index, ZEND_VM_C_GOTO(num_index_dim));
-							hval = zend_hash_func(Z_STRVAL_P(offset), Z_STRLEN_P(offset)+1);
+							if (IS_INTERNED(Z_STRVAL_P(offset))) {
+								hval = INTERNED_HASH(Z_STRVAL_P(offset));
+							} else {
+								hval = zend_hash_func(Z_STRVAL_P(offset), Z_STRLEN_P(offset)+1);
+							}
 						}
 						if (zend_hash_quick_del(ht, Z_STRVAL_P(offset), Z_STRLEN_P(offset)+1, hval) == SUCCESS &&
 					    ht == &EG(symbol_table)) {
@@ -3967,7 +3977,11 @@
 						if (!prop_dim) {
 							ZEND_HANDLE_NUMERIC_EX(Z_STRVAL_P(offset), Z_STRLEN_P(offset)+1, index, ZEND_VM_C_GOTO(num_index_prop));
 						}
-						hval = zend_hash_func(Z_STRVAL_P(offset), Z_STRLEN_P(offset)+1);
+						if (IS_INTERNED(Z_STRVAL_P(offset))) {
+							hval = INTERNED_HASH(Z_STRVAL_P(offset));
+						} else {
+							hval = zend_hash_func(Z_STRVAL_P(offset), Z_STRLEN_P(offset)+1);
+						}
 					}
 					if (zend_hash_quick_find(ht, Z_STRVAL_P(offset), Z_STRLEN_P(offset)+1, hval, (void **) &value) == SUCCESS) {
 						isset = 1;
@@ -4419,7 +4433,7 @@
 		zval_copy_ctor(&c.value);
 	}
 	c.flags = CONST_CS; /* non persistent, case sensetive */
-	c.name = zend_strndup(Z_STRVAL_P(name), Z_STRLEN_P(name));
+	c.name = IS_INTERNED(Z_STRVAL_P(name)) ? Z_STRVAL_P(name) : zend_strndup(Z_STRVAL_P(name), Z_STRLEN_P(name));
 	c.name_len = Z_STRLEN_P(name)+1;
 	c.module_number = PHP_USER_CONSTANT;
 
Index: Zend/zend_opcode.c
===================================================================
--- Zend/zend_opcode.c	(revision 35093)
+++ Zend/zend_opcode.c	(revision 35401)
@@ -243,7 +243,9 @@
 		i = op_array->last_var;
 		while (i > 0) {
 			i--;
-			efree(op_array->vars[i].name);
+			if (!IS_INTERNED(op_array->vars[i].name)) {
+				efree(op_array->vars[i].name);
+			}
 		}
 		efree(op_array->vars);
 	}
@@ -275,8 +277,10 @@
 	}
 	if (op_array->arg_info) {
 		for (i=0; i<op_array->num_args; i++) {
-			efree((char*)op_array->arg_info[i].name);
-			if (op_array->arg_info[i].class_name) {
+			if (!IS_INTERNED(op_array->arg_info[i].name)) {
+				efree((char*)op_array->arg_info[i].name);
+			}
+			if (op_array->arg_info[i].class_name && !IS_INTERNED(op_array->arg_info[i].class_name)) {
 				efree((char*)op_array->arg_info[i].class_name);
 			}
 		}
Index: main/main.c
===================================================================
--- main/main.c	(revision 35093)
+++ main/main.c	(revision 35401)
@@ -1511,6 +1511,7 @@
 	/* used to close fd's in the 3..255 range here, but it's problematic
 	 */
 	shutdown_memory_manager(1, 1 TSRMLS_CC);
+	CG(interned_strings_restore)(TSRMLS_C);
 }
 /* }}} */
 
@@ -1553,6 +1554,8 @@
 		shutdown_memory_manager(CG(unclean_shutdown), 0 TSRMLS_CC);
 	} zend_end_try();
 
+	CG(interned_strings_restore)(TSRMLS_C);
+
 	zend_try {
 		zend_unset_timeout(TSRMLS_C);
 	} zend_end_try();
@@ -1651,6 +1654,7 @@
 	zend_try {
 		shutdown_memory_manager(CG(unclean_shutdown) || !report_memleaks, 0 TSRMLS_CC);
 	} zend_end_try();
+	CG(interned_strings_restore)(TSRMLS_C);
 
 	/* 12. Reset max_execution_time */
 	zend_try {
@@ -2091,6 +2095,7 @@
 	module_startup = 0;
 
 	shutdown_memory_manager(1, 0 TSRMLS_CC);
+	CG(interned_strings_snapshot)(TSRMLS_C);
 
 	/* we're done */
 	return SUCCESS;
Index: configure.in
===================================================================
--- configure.in	(revision 297868)
+++ configure.in	(working copy)
@@ -1424,7 +1424,7 @@
     zend_list.c zend_indent.c zend_builtin_functions.c zend_sprintf.c \
     zend_ini.c zend_qsort.c zend_multibyte.c zend_ts_hash.c zend_stream.c \
     zend_iterators.c zend_interfaces.c zend_exceptions.c zend_strtod.c zend_gc.c \
-    zend_closures.c zend_float.c)
+    zend_closures.c zend_float.c zend_string.c)
 
 if test -r "$abs_srcdir/Zend/zend_objects.c"; then
   PHP_ADD_SOURCES(Zend, zend_objects.c zend_object_handlers.c zend_objects_API.c zend_default_classes.c)
Index: Zend/Makefile.am
===================================================================
--- Zend/Makefile.am	(revision 297868)
+++ Zend/Makefile.am	(working copy)
@@ -17,7 +17,7 @@
 	zend_objects_API.c zend_ts_hash.c zend_stream.c \
 	zend_default_classes.c \
 	zend_iterators.c zend_interfaces.c zend_exceptions.c \
-	zend_strtod.c zend_closures.c zend_float.c
+	zend_strtod.c zend_closures.c zend_float.c zend_string.c
 
 libZend_la_LDFLAGS =
 libZend_la_LIBADD = @ZEND_EXTRA_LIBS@
Index: win32/build/config.w32
===================================================================
--- win32/build/config.w32	(revision 297868)
+++ win32/build/config.w32	(working copy)
@@ -325,7 +325,7 @@
 	zend_stream.c zend_iterators.c zend_interfaces.c zend_objects.c \
 	zend_object_handlers.c zend_objects_API.c \
 	zend_default_classes.c zend_execute.c zend_strtod.c zend_gc.c zend_closures.c \
-	zend_float.c");
+	zend_float.c zend_string.c");
 
 if (VCVERS == 1200) {
 	AC_DEFINE('ZEND_DVAL_TO_LVAL_CAST_OK', 1);
Index: Zend/zend_string.c
===================================================================
--- Zend/zend_string.c	(revision 0)
+++ Zend/zend_string.c	(revision 0)
@@ -0,0 +1,229 @@
+/*
+   +----------------------------------------------------------------------+
+   | Zend Engine                                                          |
+   +----------------------------------------------------------------------+
+   | Copyright (c) 1998-2010 Zend Technologies Ltd. (http://www.zend.com) |
+   +----------------------------------------------------------------------+
+   | This source file is subject to version 2.00 of the Zend license,     |
+   | that is bundled with this package in the file LICENSE, and is        | 
+   | available through the world-wide-web at the following url:           |
+   | http://www.zend.com/license/2_00.txt.                                |
+   | If you did not receive a copy of the Zend license and are unable to  |
+   | obtain it through the world-wide-web, please send a note to          |
+   | license@zend.com so we can mail you a copy immediately.              |
+   +----------------------------------------------------------------------+
+   | Authors: Dmitry Stogov <dmitry@zend.com>                             |
+   +----------------------------------------------------------------------+
+*/
+
+/* $Id: $ */
+
+#include "zend.h"
+#include "zend_globals.h"
+
+#if ZEND_DEBUG && 1
+# define ZEND_DEBUG_INTERNED_STRINGS 1
+#endif
+
+#if ZEND_DEBUG_INTERNED_STRINGS
+# include <sys/mman.h>
+#endif
+
+static char *zend_new_interned_string(char *str, int len, int free_src TSRMLS_DC);
+static void zend_interned_strings_snapshot(TSRMLS_D);
+static void zend_interned_strings_restore(TSRMLS_D);
+
+void zend_interned_strings_init(TSRMLS_D)
+{
+#ifndef ZTS
+	size_t size = 1024 * 1024;
+
+#if ZEND_DEBUG_INTERNED_STRINGS
+	CG(interned_strings_start) = valloc(size);
+#else
+	CG(interned_strings_start) = malloc(size);
+#endif
+
+	CG(interned_strings_top) = CG(interned_strings_start);
+	CG(interned_strings_snapshot_top) = CG(interned_strings_start);
+	CG(interned_strings_end) = CG(interned_strings_start) + size;
+
+	zend_hash_init(&CG(interned_strings), 0, NULL, NULL, 1);
+	
+	CG(interned_strings).nTableMask = CG(interned_strings).nTableSize - 1;
+	CG(interned_strings).arBuckets = (Bucket **) pecalloc(CG(interned_strings).nTableSize, sizeof(Bucket *), CG(interned_strings).persistent);
+
+#if ZEND_DEBUG_INTERNED_STRINGS
+	mprotect(CG(interned_strings_start), CG(interned_strings_end) - CG(interned_strings_start), PROT_READ);
+#endif
+
+#endif
+
+	CG(new_interned_string) = zend_new_interned_string;
+	CG(interned_strings_snapshot) = zend_interned_strings_snapshot;
+	CG(interned_strings_restore) = zend_interned_strings_restore;
+}
+
+void zend_interned_strings_dtor(TSRMLS_D)
+{
+#ifndef ZTS
+#if ZEND_DEBUG_INTERNED_STRINGS
+	mprotect(CG(interned_strings_start), CG(interned_strings_end) - CG(interned_strings_start), PROT_WRITE | PROT_READ);
+#endif
+	free(CG(interned_strings).arBuckets);
+	free(CG(interned_strings_start));
+#endif
+}
+
+static char *zend_new_interned_string(char *arKey, int nKeyLength, int free_src TSRMLS_DC)
+{
+#ifndef ZTS
+	ulong h;
+	uint nIndex;
+	Bucket *p;
+
+	if (IS_INTERNED(arKey)) {
+		return arKey;
+	}
+
+	h = zend_inline_hash_func(arKey, nKeyLength);
+	nIndex = h & CG(interned_strings).nTableMask;
+	p = CG(interned_strings).arBuckets[nIndex];
+	while (p != NULL) {
+		if ((p->h == h) && (p->nKeyLength == nKeyLength)) {
+			if (!memcmp(p->arKey, arKey, nKeyLength)) {
+				if (free_src) {
+					efree(arKey);
+				}
+				return p->arKey;
+			}
+		}
+		p = p->pNext;
+	}
+	
+	if (CG(interned_strings_top) + ZEND_MM_ALIGNED_SIZE(sizeof(Bucket) + nKeyLength) >=
+	    CG(interned_strings_end)) {
+	    /* no memory */
+		return arKey;
+	}
+
+	p = (Bucket *) CG(interned_strings_top);
+	CG(interned_strings_top) += ZEND_MM_ALIGNED_SIZE(sizeof(Bucket) + nKeyLength);
+
+#if ZEND_DEBUG_INTERNED_STRINGS
+	mprotect(CG(interned_strings_start), CG(interned_strings_end) - CG(interned_strings_start), PROT_READ | PROT_WRITE);
+#endif
+	
+	p->arKey = (char*)(p+1);
+	memcpy(p->arKey, arKey, nKeyLength);
+	if (free_src) {
+		efree(arKey);
+	}
+	p->nKeyLength = nKeyLength;
+	p->h = h;
+	p->pData = &p->pDataPtr;
+	p->pDataPtr = p;
+	
+	p->pNext = CG(interned_strings).arBuckets[nIndex];
+	p->pLast = NULL;
+	if (p->pNext) {
+		p->pNext->pLast = p;
+	}
+
+	HANDLE_BLOCK_INTERRUPTIONS();
+	
+	p->pListLast = CG(interned_strings).pListTail;
+	CG(interned_strings).pListTail = p;
+	p->pListNext = NULL;
+	if (p->pListLast != NULL) {
+		p->pListLast->pListNext = p;
+	}
+	if (!CG(interned_strings).pListHead) {
+		CG(interned_strings).pListHead = p;
+	}
+
+	CG(interned_strings).arBuckets[nIndex] = p;
+
+	HANDLE_UNBLOCK_INTERRUPTIONS();
+
+	CG(interned_strings).nNumOfElements++;
+
+	if (CG(interned_strings).nNumOfElements > CG(interned_strings).nTableSize) {
+		if ((CG(interned_strings).nTableSize << 1) > 0) {	/* Let's double the table size */
+			Bucket **t = (Bucket **) perealloc_recoverable(CG(interned_strings).arBuckets, (CG(interned_strings).nTableSize << 1) * sizeof(Bucket *), CG(interned_strings).persistent);
+
+			if (t) {
+				HANDLE_BLOCK_INTERRUPTIONS();
+				CG(interned_strings).arBuckets = t;
+				CG(interned_strings).nTableSize = (CG(interned_strings).nTableSize << 1);
+				CG(interned_strings).nTableMask = CG(interned_strings).nTableSize - 1;
+				zend_hash_rehash(&CG(interned_strings));
+				HANDLE_UNBLOCK_INTERRUPTIONS();
+			}
+		}
+	}
+
+#if ZEND_DEBUG_INTERNED_STRINGS
+	mprotect(CG(interned_strings_start), CG(interned_strings_end) - CG(interned_strings_start), PROT_READ);
+#endif
+
+	return p->arKey;
+#else
+	return arKey;
+#endif
+}
+
+static void zend_interned_strings_snapshot(TSRMLS_D)
+{
+	CG(interned_strings_snapshot_top) = CG(interned_strings_top);
+}
+
+static void zend_interned_strings_restore(TSRMLS_D)
+{
+#ifndef ZTS
+	Bucket *p;
+	int i;
+#endif
+
+	CG(interned_strings_top) = CG(interned_strings_snapshot_top);
+
+#ifndef ZTS
+#if ZEND_DEBUG_INTERNED_STRINGS
+	mprotect(CG(interned_strings_start), CG(interned_strings_end) - CG(interned_strings_start), PROT_WRITE | PROT_READ);
+#endif
+
+	for (i = 0; i < CG(interned_strings).nTableSize; i++) {
+		p = CG(interned_strings).arBuckets[i];
+		while (p && p->arKey > CG(interned_strings_top)) {
+			CG(interned_strings).nNumOfElements--;
+			if (p->pListLast != NULL) {
+				p->pListLast->pListNext = p->pListNext;
+			} else { 
+				CG(interned_strings).pListHead = p->pListNext;
+			}
+			if (p->pListNext != NULL) {
+				p->pListNext->pListLast = p->pListLast;
+			} else {
+				CG(interned_strings).pListTail = p->pListLast;
+			}
+			p = p->pNext;
+		}
+		if (p) {
+			p->pLast = NULL;
+		}
+		CG(interned_strings).arBuckets[i] = p;
+	}
+
+#if ZEND_DEBUG_INTERNED_STRINGS
+	mprotect(CG(interned_strings_start), CG(interned_strings_end) - CG(interned_strings_start), PROT_READ);
+#endif
+#endif
+}
+
+/*
+ * Local variables:
+ * tab-width: 4
+ * c-basic-offset: 4
+ * indent-tabs-mode: t
+ * End:
+ */
Index: Zend/zend_string.h
===================================================================
--- Zend/zend_string.h	(revision 0)
+++ Zend/zend_string.h	(revision 0)
@@ -0,0 +1,55 @@
+/*
+   +----------------------------------------------------------------------+
+   | Zend Engine                                                          |
+   +----------------------------------------------------------------------+
+   | Copyright (c) 1998-2010 Zend Technologies Ltd. (http://www.zend.com) |
+   +----------------------------------------------------------------------+
+   | This source file is subject to version 2.00 of the Zend license,     |
+   | that is bundled with this package in the file LICENSE, and is        | 
+   | available through the world-wide-web at the following url:           |
+   | http://www.zend.com/license/2_00.txt.                                |
+   | If you did not receive a copy of the Zend license and are unable to  |
+   | obtain it through the world-wide-web, please send a note to          |
+   | license@zend.com so we can mail you a copy immediately.              |
+   +----------------------------------------------------------------------+
+   | Authors: Dmitry Stogov <dmitry@zend.com>                             |
+   +----------------------------------------------------------------------+
+*/
+
+/* $Id: $ */
+
+#ifndef ZEND_STRING_H
+#define ZEND_STRING_H
+
+#include "zend.h"
+
+void zend_interned_strings_init(TSRMLS_D);
+void zend_interned_strings_dtor(TSRMLS_D);
+
+#ifndef ZTS
+
+#define IS_INTERNED(s) \
+	(((s) >= CG(interned_strings_start)) && ((s) < CG(interned_strings_end)))
+
+#else
+
+#define IS_INTERNED(s) \
+	(0)
+
+#endif
+
+#define INTERNED_LEN(s) \
+	(((Bucket*)(((char*)(s))-sizeof(Bucket)))->nKeyLength)
+
+#define INTERNED_HASH(s) \
+	(((Bucket*)(((char*)(s))-sizeof(Bucket)))->h)
+
+#endif /* ZEND_STRING_H */
+
+/*
+ * Local variables:
+ * tab-width: 4
+ * c-basic-offset: 4
+ * indent-tabs-mode: t
+ * End:
+ */